package com.sun.org.apache.bcel.internal.classfile;

import java.io.DataOutputStream;
import com.sun.org.apache.bcel.internal.Const;
import java.io.IOException;
import java.io.DataInput;

public class ConstantPool implements Cloneable, Node
{
    private Constant[] constant_pool;
    
    public ConstantPool(Constant[] constant_pool) {
        this.constant_pool = constant_pool;
    }
    
    public ConstantPool(DataInput input) throws IOException, ClassFormatException {
        int constant_pool_count = input.readUnsignedShort();
        this.constant_pool = new Constant[constant_pool_count];
        for (int i = 1; i < constant_pool_count; ++i) {
            this.constant_pool[i] = Constant.readConstant(input);
            byte tag = this.constant_pool[i].getTag();
            if (tag == 6 || tag == 5) {
                ++i;
            }
        }
    }
    
    @Override
    public void accept(Visitor v) {
        v.visitConstantPool(this);
    }
    
    public String constantToString(Constant c) throws ClassFormatException {
        byte tag = c.getTag();
        String str;
        switch (tag) {
            case 7: {
                int i = ((ConstantClass)c).getNameIndex();
                c = this.getConstant(i, (byte)1);
                str = Utility.compactClassName(((ConstantUtf8)c).getBytes(), false);
                break;
            }
            case 8: {
                int i = ((ConstantString)c).getStringIndex();
                c = this.getConstant(i, (byte)1);
                str = "\"" + escape(((ConstantUtf8)c).getBytes()) + "\"";
                break;
            }
            case 1: {
                str = ((ConstantUtf8)c).getBytes();
                break;
            }
            case 6: {
                str = String.valueOf(((ConstantDouble)c).getBytes());
                break;
            }
            case 4: {
                str = String.valueOf(((ConstantFloat)c).getBytes());
                break;
            }
            case 5: {
                str = String.valueOf(((ConstantLong)c).getBytes());
                break;
            }
            case 3: {
                str = String.valueOf(((ConstantInteger)c).getBytes());
                break;
            }
            case 12: {
                str = this.constantToString(((ConstantNameAndType)c).getNameIndex(), (byte)1) + " " + this.constantToString(((ConstantNameAndType)c).getSignatureIndex(), (byte)1);
                break;
            }
            case 9:
            case 10:
            case 11: {
                str = this.constantToString(((ConstantCP)c).getClassIndex(), (byte)7) + "." + this.constantToString(((ConstantCP)c).getNameAndTypeIndex(), (byte)12);
                break;
            }
            case 15: {
                ConstantMethodHandle cmh = (ConstantMethodHandle)c;
                str = Const.getMethodHandleName(cmh.getReferenceKind()) + " " + this.constantToString(cmh.getReferenceIndex(), this.getConstant(cmh.getReferenceIndex()).getTag());
                break;
            }
            case 16: {
                ConstantMethodType cmt = (ConstantMethodType)c;
                str = this.constantToString(cmt.getDescriptorIndex(), (byte)1);
                break;
            }
            case 18: {
                ConstantInvokeDynamic cid = (ConstantInvokeDynamic)c;
                str = cid.getBootstrapMethodAttrIndex() + ":" + this.constantToString(cid.getNameAndTypeIndex(), (byte)12);
                break;
            }
            default: {
                throw new RuntimeException("Unknown constant type " + (int)tag);
            }
        }
        return str;
    }
    
    private static String escape(String str) {
        int len = str.length();
        StringBuilder buf = new StringBuilder(len + 5);
        char[] ch = str.toCharArray();
        for (int i = 0; i < len; ++i) {
            switch (ch[i]) {
                case '\n': {
                    buf.append("\\n");
                    break;
                }
                case '\r': {
                    buf.append("\\r");
                    break;
                }
                case '\t': {
                    buf.append("\\t");
                    break;
                }
                case '\b': {
                    buf.append("\\b");
                    break;
                }
                case '\"': {
                    buf.append("\\\"");
                    break;
                }
                default: {
                    buf.append(ch[i]);
                    break;
                }
            }
        }
        return buf.toString();
    }
    
    public String constantToString(int index, byte tag) throws ClassFormatException {
        Constant c = this.getConstant(index, tag);
        return this.constantToString(c);
    }
    
    public void dump(DataOutputStream file) throws IOException {
        file.writeShort(this.constant_pool.length);
        for (int i = 1; i < this.constant_pool.length; ++i) {
            if (this.constant_pool[i] != null) {
                this.constant_pool[i].dump(file);
            }
        }
    }
    
    public Constant getConstant(int index) {
        if (index >= this.constant_pool.length || index < 0) {
            throw new ClassFormatException("Invalid constant pool reference: " + index + ". Constant pool size is: " + this.constant_pool.length);
        }
        return this.constant_pool[index];
    }
    
    public Constant getConstant(int index, byte tag) throws ClassFormatException {
        Constant c = this.getConstant(index);
        if (c == null) {
            throw new ClassFormatException("Constant pool at index " + index + " is null.");
        }
        if (c.getTag() != tag) {
            throw new ClassFormatException("Expected class `" + Const.getConstantName((int)tag) + "' at index " + index + " and got " + c);
        }
        return c;
    }
    
    public Constant[] getConstantPool() {
        return this.constant_pool;
    }
    
    public String getConstantString(int index, byte tag) throws ClassFormatException {
        Constant c = this.getConstant(index, tag);
        int i;
        switch (tag) {
            case 7: {
                i = ((ConstantClass)c).getNameIndex();
                break;
            }
            case 8: {
                i = ((ConstantString)c).getStringIndex();
                break;
            }
            default: {
                throw new RuntimeException("getConstantString called with illegal tag " + (int)tag);
            }
        }
        c = this.getConstant(i, (byte)1);
        return ((ConstantUtf8)c).getBytes();
    }
    
    public int getLength() {
        return (this.constant_pool == null) ? 0 : this.constant_pool.length;
    }
    
    public void setConstant(int index, Constant constant) {
        this.constant_pool[index] = constant;
    }
    
    public void setConstantPool(Constant[] constant_pool) {
        this.constant_pool = constant_pool;
    }
    
    @Override
    public String toString() {
        StringBuilder buf = new StringBuilder();
        for (int i = 1; i < this.constant_pool.length; ++i) {
            buf.append(i).append(")").append(this.constant_pool[i]).append("\n");
        }
        return buf.toString();
    }
    
    public ConstantPool copy() {
        ConstantPool c = null;
        try {
            c = (ConstantPool)this.clone();
            c.constant_pool = new Constant[this.constant_pool.length];
            for (int i = 1; i < this.constant_pool.length; ++i) {
                if (this.constant_pool[i] != null) {
                    c.constant_pool[i] = this.constant_pool[i].copy();
                }
            }
        }
        catch (CloneNotSupportedException ex) {}
        return c;
    }
}
